unit U.LazyLoader;

interface

uses System.TypInfo, U.LazyLoader.Interfaces;

type

  TLazyLoader = class(TInterfacedObject, ILazyLoader)
  private
    FInternalInstance: IInterface;
    FTypeInfo: PTypeInfo;
    FID: Integer;
    procedure CreateInternalInstance;
  public
    constructor Create(const ATypeInfo:PTypeInfo; const AID:Integer);
    function GetValue: IInterface;
  end;

implementation

uses iORM, System.SysUtils;

{ TLazyLoader }

constructor TLazyLoader.Create(const ATypeInfo: PTypeInfo; const AID: Integer);
begin
  FInternalInstance := nil;
  FTypeInfo := ATypeInfo;
  FID := AID;
end;

procedure TLazyLoader.CreateInternalInstance;
var
  LObj: TObject;
begin
  LObj := io.Load(FTypeInfo).ByOID(FID).ToObject;
  if not Supports(LObj, IInterface, FInternalInstance) then
    raise Exception.Create('The instance is not an interface implementer.');
end;

function TLazyLoader.GetValue: IInterface;
begin
  if not Assigned(FInternalInstance) then
    CreateInternalInstance;
  Result := FInternalInstance;
end;

end.
