unit U.LazyLoader;

interface

uses System.TypInfo, System.Rtti;

type

{$region 'Step 3'}
//  TLazyLoader<T:IInterface> = class(TInterfacedObject, ILazyLoader<T>)
//  private
//    FInternalInstance: T;
//    FID: Integer;
//    procedure CreateInternalInstance;
//  public
//    constructor Create(const AID:Integer);
//    function Invoke: T;  // Automatically invoked!!!
//  end;
{$endregion}

  TLazyLoader = class(TVirtualInterface)
  private
    FInternalInstance: IInterface;
    FTypeInfo: PTypeInfo;
    FID: Integer;
    function GetInternalInstance: IInterface;
    procedure DoInvoke(Method: TRttiMethod; const Args: TArray<TValue>; out Result: TValue);
  public
    constructor Create(PIID: PTypeInfo; AID: Integer); overload;
  end;

implementation

uses iORM, System.SysUtils;

{ TLazyLoader }

{$region 'Step 3'}
//  constructor TLazyLoader<T>.Create(const AID: Integer);
//  begin
//    FInternalInstance := nil;
//    FID := AID;
//  end;
{$endregion}

constructor TLazyLoader.Create(PIID: PTypeInfo; AID: Integer);
begin
  inherited Create(PIID, DoInvoke);
  FTypeInfo := PIID;
  FInternalInstance := nil;
  FID := AID;
end;

{$region 'Step 3'}
//  procedure TLazyLoader<T>.CreateInternalInstance;
//  begin
//    FInternalInstance := io.Load<T>.ByOID(FID).ToObject;
//  end;
{$endregion}

function TLazyLoader.GetInternalInstance: IInterface;
var
  LObj: TObject;
begin
  if not Assigned(FInternalInstance) then
  begin
    LObj := io.Load(FTypeInfo).ByOID(FID).ToObject;
    if not Supports(LObj, IInterface, FInternalInstance) then
      raise Exception.Create('The instance is not an interface implementer.');
  end;
  Result := FInternalInstance;
end;

{$region 'Step 3'}
//  //Step 2:  function TLazyLoader<T>.GetValue: T;
//  function TLazyLoader<T>.Invoke: T;
//  begin
//    if not Assigned(FInternalInstance) then
//      CreateInternalInstance;
//    Result := FInternalInstance;
//  end;
{$endregion}

procedure TLazyLoader.DoInvoke(Method: TRttiMethod; const Args: TArray<TValue>; out Result: TValue);
var
  LContext: TRttiContext;
  LType: TRttiType;
  LDestMethod: TRttiMethod;
  LArgs: TArray<TValue>;
  LObj: TObject;
begin
  LObj := GetInternalInstance as TObject;
  // Get the destination RttiMethod
  LContext := TRttiContext.Create;
  LType := LContext.GetType(LObj.ClassInfo);
  LDestMethod := LType.GetMethod(Method.Name);
  // Do not consider the first element of Args
  LArgs := Copy(Args, 1, Length(Args));
  // Invoke the method
  Result := LDestMethod.Invoke(LObj, LArgs);
end;

end.
