unit U.LazyLoader;

interface

uses System.TypInfo, System.Rtti;

type

{$region 'Step 4'}
//  TLazyLoader = class(TVirtualInterface)
//  private
//    FInternalInstance: IInterface;
//    FTypeInfo: PTypeInfo;
//    FID: Integer;
//    function GetInternalInstance: IInterface;
//    procedure DoInvoke(Method: TRttiMethod; const Args: TArray<TValue>; out Result: TValue);
//  public
//    constructor Create(PIID: PTypeInfo; AID: Integer); overload;
//  end;
{$endregion}

  TLazyLoader<T:IInterface> = class(TVirtualInterface)
  private
    FInternalInstance: T;
    FID: Integer;
    function GetInternalInstance: T;
    procedure DoInvoke(Method: TRttiMethod; const Args: TArray<TValue>; out Result: TValue);
  public
    constructor Create(AID: Integer); overload;
//Step 4: constructor Create(PIID: PTypeInfo; AID: Integer); overload;
  end;

implementation

uses iORM, System.SysUtils;

{ TLazyLoader }

{$region 'Step 4'}
//  constructor TLazyLoader.Create(PIID: PTypeInfo; AID: Integer);
//  begin
//    inherited Create(PIID, DoInvoke);
//    FTypeInfo := PIID;
//    FInternalInstance := nil;
//    FID := AID;
//  end;
{$endregion}

constructor TLazyLoader<T>.Create(AID: Integer);
begin
  inherited Create(TypeInfo(T), DoInvoke);
  FInternalInstance := nil;
  FID := AID;
end;

{$region 'Step 4'}
//  function TLazyLoader.GetInternalInstance: IInterface;
//  var
//    LObj: TObject;
//  begin
//    if not Assigned(FInternalInstance) then
//    begin
//      LObj := io.Load(FTypeInfo).ByOID(FID).ToObject;
//      if not Supports(LObj, IInterface, FInternalInstance) then
//        raise Exception.Create('The instance is not an interface implementer.');
//    end;
//    Result := FInternalInstance;
//  end;
{$endregion}

function TLazyLoader<T>.GetInternalInstance: T;
begin
  if not Assigned(FInternalInstance) then
    FInternalInstance := io.Load<T>.ByOID(FID).ToObject;
  Result := FInternalInstance;
end;

procedure TLazyLoader<T>.DoInvoke(Method: TRttiMethod; const Args: TArray<TValue>; out Result: TValue);
var
  LContext: TRttiContext;
  LType: TRttiType;
  LDestMethod: TRttiMethod;
  LArgs: TArray<TValue>;
  LObj: TObject;
begin
  LObj := GetInternalInstance as TObject;
  // Get the destination RttiMethod
  LContext := TRttiContext.Create;
  LType := LContext.GetType(LObj.ClassInfo);
  LDestMethod := LType.GetMethod(Method.Name);
  // Do not consider the first element of Args
  LArgs := Copy(Args, 1, Length(Args));
  // Invoke the method
  Result := LDestMethod.Invoke(LObj, LArgs);
end;

end.
